from __future__ import print_function
from vcdvcd import VCDVCD

import math
import os
import sys
import numpy
import csv
import pickle
from sklearn.preprocessing import scale
from scipy.spatial import distance
import secrets
from scipy.stats import ttest_ind


class PowerAnalyzePersistent(object):

    def __init__(self, project_dir,name, num_plaintext, vcdpath='', timeperiod=None, vcdcatpath='vcdcat', multiple_vcd=False):
        self.project_dir=project_dir
        self.name = name
        self.num_plaintext = num_plaintext
        self.vcdcatpath = os.path.abspath(vcdcatpath)
        self.vcdpath=vcdpath
        self.timeperiod=timeperiod
        self.multiple_vcd = multiple_vcd
        self.module_map = {}
        self.signal_map = {}

        #create appropriate Map direcctories
        self.maps_path = os.path.join(self.project_dir,'maps_'+str(self.num_plaintext))
        self.value_map_path = os.path.join(self.maps_path, self.name+'_value_map')
        self.SA_map_path = os.path.join(self.maps_path, self.name+'_SA_map')
        self.create_map_dir()

        #Initialize lists for the value map and SA map
        self.value_map = []
        self.SA_map = []

    def create_map_dir(self):
        try:
            os.mkdir(self.maps_path)
        except FileExistsError:
            if self.multiple_vcd:
                pass
            else:
                print("Map directory already exists")

        try:
            os.mkdir(self.value_map_path)
        except FileExistsError:
            if self.multiple_vcd:
                pass
            else:
                print("Value Map directory already exists")
            #shutil.rmtree(os.path.join(project_dir,'Traces_HW_'+str(num_pt)))
            #os.mkdir(os.path.join(project_dir,'Traces_HW_'+str(num_pt)))

        try:
            os.mkdir(self.SA_map_path)
        except FileExistsError:
            if self.multiple_vcd:
                pass
            else:
                print("SA Map directory already exists")


    def set_vcdPath(self, vcdpath):
        self.vcdpath = vcdpath

    def set_timeperiod(self, timeperiod):
        self.timeperiod = timeperiod


    def create_VCDDumpMap(self):
        if not self.vcdpath:
            raise Exception("VCD path not set")

        #check if the module maps already exists

        # if os.path.exists(os.path.join(self.maps_path, self.name+'_module_map.dict')):
        #     print('Module Map already exists')
        #     print('Reading Maps...')
        #     self.read_maps()
        #     return

        if os.path.exists(self.vcdpath+".txt"):
            print("VCD dump.txt already exists")
        else:
            print("VCD dump do not exist... generating it")
            command = self.vcdcatpath + " " + os.path.join(self.vcdpath) + " > " + os.path.join(self.vcdpath + ".txt")
            os.system(command)

        #vcd = VCDVCD(self.vcdpath,print_deltas=False, print_dumps=True,signals=[],store_tvs=False)

        print("VCD dump.txt generation complete... parsing the txt file")


        with open(self.vcdpath+".txt",'r') as fd:
            self.data = fd.read().splitlines()

        for _ in range(0,len(self.data)):
            if self.data[0] == '':
                del self.data[0] #delete this line
                #cols =  self.data.pop(0).split() #Read the column map and create a list in valuemap
            elif self.data[0][0] == '=':
                del self.data[0] #delete the line containing "==="
                self.create_valuemap()
                break
            else:
                line = self.data.pop(0)
                line = line.split()
                index = line[0]
                module_signal = line[1].split('.')

                if len(module_signal) > 1:
                    #update the signal map     
                    self.signal_map[line[1]] = index

                    #update the module map with signals
                    for i in range(len(module_signal)): #iterate over every module and signal in the line
                        #consider only if it is not the last wire
                        if i < len(module_signal)-1: 
                            
                            if module_signal[i] not in self.module_map.keys():
                                self.module_map[module_signal[i]] = []
                                self.module_map[module_signal[i]].append(line[1])
                            else:
                                #module key already exists 
                                self.module_map[module_signal[i]].append(line[1])
                                    

        #create the SA Map
        self.create_SAmap()

        #Save the Maps
        self.save_maps()

        #Save Value and SA Maps
        self.save_value_maps()

    def create_valuemap(self):
        print("Creating Value Map")
        self.value_map = [ [] for _ in range(len(list(self.signal_map.keys()))+1 )]
        length = len(self.data)
        for _ in range(length):
            line = self.data.pop(0)
            line = line.replace("x", "0")
            line = line.replace("z","0")
            values = line.split()
            for i,val in enumerate(values):
                self.value_map[i].append(val)

    def create_SAmap(self):
        if not self.value_map:
            raise Exception("Value map dictionary not created")
        
        print("Creating SA Map")
        self.SA_map = [ [] for _ in range(len(self.value_map))]

        for id in range(len(self.value_map)):       
            values = self.value_map[id]
            self.SA_map[id] = list(map(lambda i: '{:x}'.format(int(values[i-1], 16) ^ int(values[i], 16)) if i>0 else values[i], range(0,len(values)) ))


    def save_maps(self):
        if not (self.module_map or self.signal_map or self.value_map or self.SA_map):
            raise Exception("Dictionary mappings not set")
        
        print("Saving all the maps")

        file = open(os.path.join(self.maps_path, self.name+'_module_map.dict'), 'wb') 
        pickle.dump(self.module_map, file) 
        file.close() 

        file = open(os.path.join(self.maps_path,  self.name+'_signal_map.dict'), 'wb') 
        pickle.dump(self.signal_map, file) 
        file.close() 

        

    def save_value_maps(self):

        for i in range(len(self.value_map)):
            with open( os.path.join(self.value_map_path, 'value_map_'+str(i)),"a") as f:
                f.write("\n".join(self.value_map[i])) 

        for i in range(len(self.SA_map)):
            with open( os.path.join(self.SA_map_path, 'SA_map_'+str(i)),"a") as f:
                f.write("\n".join(self.SA_map[i]))

        self.value_map.clear()
        self.SA_map.clear()


    def read_maps(self):
        """
        Read all the maps stored in the project directory
        """
        # if not self.vcdpath:
        #     raise Exception("VCD path not set")

        print("Reading Module and Signal map for "+self.name)

        with open(os.path.join(self.maps_path, self.name+'_module_map.dict'), 'rb') as handle:
            self.module_map = pickle.load(handle)
        
        with open(os.path.join(self.maps_path,  self.name+'_signal_map.dict'), 'rb') as handle:
            self.signal_map = pickle.load(handle)


    def get_maps(self):
        return self.module_map, self.signal_map, self.value_map, self.SA_map

    def get_HWtraces(self, module_name, num_plaintext=None):
        """
        Takes the module name and for that module gives the power trace based on the Hamming weight.
        To reduce the memory consumption only lists for the specific signals in the module are read from the harddisk maps
        """
        print("Getting HW Power Trace for module " + module_name)

        if not self.timeperiod:
            raise Exception("Time period of the power traces not set")

        #if the number of plaintexts is not provided it works on the default number of plaintexts for the power traces
        if num_plaintext==None:
            num_plaintext=self.num_plaintext

        #self.HW is list of lists where each list corresponds to power trace for 1 plaintext
        self.HW = [ [] for _ in range(num_plaintext)]
        
        signals = self.module_map[module_name]
        
        for i,signal in enumerate(signals):
            if i==0:
                with open(os.path.join(self.value_map_path, 'value_map_'+self.signal_map[signal] )) as fd:
                    l = fd.read().splitlines()
            else:
                with open(os.path.join(self.value_map_path, 'value_map_'+self.signal_map[signal] )) as fd:
                    temp = fd.read().splitlines()
                l = list(map(str.__add__,l, temp))
        
        #if len(l[-1]) != len(l[0]):
        #    del l[-1] #delete the last trace if it is not of the same length

        key = 0 #Not a dictionary key
        i=0

        for line in l:
            hw = self.hex2bin(line).count('1')
            i = i+1
            self.HW[key].append(hw)

            if i == self.timeperiod:
                key = key+1
                i = 0
            if key == num_plaintext:
                break

        if len(self.HW[key-1]) != len(self.HW[0]): #If last trace not of same length
            self.HW.pop(key-1) 

        return self.HW
    
    def get_SAtraces(self,module_name,clean=False, num_plaintext=None, clock_trigger='positive', clock_start='0'):
        
        print("Getting SA Power Trace for module " + module_name)

        if not self.timeperiod:
            raise Exception("Time period of the power traces not set")
        
        if num_plaintext==None:
            num_plaintext=self.num_plaintext
        
        SA_trace = [ [] for _ in range(num_plaintext)]
        
        signals = self.module_map[module_name]

        for i,signal in enumerate(signals):
            if i==0:
                with open(os.path.join(self.SA_map_path, 'SA_map_'+self.signal_map[signal] )) as fd:
                    l = fd.read().splitlines()
            else:
                with open(os.path.join(self.SA_map_path, 'SA_map_'+self.signal_map[signal] )) as fd:
                    temp = fd.read().splitlines()
                l = list(map(str.__add__,l, temp))

        key = 0
        i=0
        
        for line in l:
            SA = self.hex2bin(line).count('1')

            if i!=0 and i%self.timeperiod==0:
                key += 1
                if key == num_plaintext:
                    break

                
            #Following condition is needed to align the traces in case of multiple VCDs
            #self.num_plaintext and num_plaintext are two different things in this module
            if key!=0 and key%self.num_plaintext==0 and i%self.timeperiod==0:
                 SA_trace[key].append(0)
                 i += 1


            SA_trace[key].append(SA)
            i += 1

            
        del(l)
        return SA_trace

    
    @staticmethod
    def KL_Divergence(trace1, trace2):
        #Check the length of last trace and if not equal remove it
        if len(trace1[-1]) != len(trace1[0]):
            trace1.pop(-1)
            trace2.pop(-1)
        trace1 = numpy.array(trace1)
        trace2 = numpy.array(trace2)
        mean_std1 = PowerAnalyzePersistent.get_traces_mean_std(trace1)
        mean_std2 = PowerAnalyzePersistent.get_traces_mean_std(trace2)

        multiKL_divergence = lambda a,b,c,d: map(PowerAnalyzePersistent.klDivergence, a,b,c,d)
        kl = list(multiKL_divergence(mean_std1['mean'], mean_std2['mean'], mean_std1['std'], mean_std2['std']))
        return kl
    

    @staticmethod
    def t_test(trace1,trace2):
        t_values = []
        p_values = []
        for col1,col2 in zip(numpy.transpose(trace1), numpy.transpose(trace2)):
            t,p = ttest_ind(col1,col2, equal_var=False)
            t_values.append(t)
            p_values.append(p)
        return t_values,p_values


    @staticmethod
    def Jensen_Shannon(trace1, trace2):
        js=[]
        for col1,col2 in zip(numpy.transpose(trace1), numpy.transpose(trace2)):
            js.append(distance.jensenshannon(col1,col2))
        return js

    @staticmethod
    def get_traces_mean_std(traces_array, axis=0):
        """
        Returns the dictionary for mean and std of the 2d array.
        Default axis is columnwise.
        """

        mean_std = {}
        mean_std['mean'] = list(numpy.mean(traces_array, axis=axis))
        mean_std['std'] = list(numpy.std(traces_array, axis=axis))
        return mean_std

    @staticmethod
    def klDivergence(x_mu, y_mu, x_std, y_std):
        from math import log
        try:
            kl1 = (((x_mu-y_mu)**2 + x_std**2 - y_std**2)/(2*(y_std**2))) + log(y_std/x_std)
        except:
            kl1 = 'nan'
        
        try:
            kl2 = (((y_mu-x_mu)**2 + y_std**2 - x_std**2)/(2*(x_std**2))) + log(x_std/y_std)
        except:
            kl2 = 'nan'
        
        if kl1 == 'nan' and kl2 == 'nan':
            return 'nan'
        elif kl1 != 'nan' and kl2 != 'nan':
            return max(kl1,kl2)
        else:
            if kl1 == 'nan':
                return kl2
            else:
                return kl1

    @staticmethod
    def save_metric_dict(m_dict, filename):
        keys = sorted(m_dict.keys())
        with open(filename, 'w') as outfile:
            writer = csv.writer(outfile, delimiter = ",")
            writer.writerow(keys)
            writer.writerows(zip(*[m_dict[key] for key in keys]))

    @staticmethod
    def GenerateKeyPairs(index):
        Keypairs = {1: (secrets.token_hex(16), secrets.token_hex(16)),
                    2:("00"*16, "FF"*16),
                    3: ("11"*16, "EE"*16),
                    4: ("33"*16, "CC"*16),
                    5: ("77"*16, "88"*16)}

        Keyvalues = Keypairs.get(index,None)
        return Keyvalues


    def hex2bin(self, str):
        bin = ['0000','0001','0010','0011',
                '0100','0101','0110','0111',
                '1000','1001','1010','1011',
                '1100','1101','1110','1111']
        aa = ''
        for i in range(len(str)):
            aa += bin[int(str[i],base=16)]
        return aa